from typing import cast
import json
import logging

from pytractions.base import TDict, doc

from .quay_client import QuayClient, ManifestTypeError, ManifestNotFoundError

from .types import ManifestList, Manifest


LOG = logging.getLogger()
logging.basicConfig()
LOG.setLevel(logging.INFO)


class FakeQuayClient(QuayClient):
    """Class for performing Docker HTTP API operations with the Quay registry."""

    fake_manifests: TDict[str, TDict[str, str]]
    d_fake_manifests: str = doc("Fake manifests for testing.")

    def __post_init__(self):
        """Fake quay client post init."""
        LOG.info("Populate fake manifests:")
        for image, media_types in self.fake_manifests:
            LOG.info("Image: {}".format(image))
            for mtype in media_types:
                LOG.info("\tMedia Type {}".format(mtype))

    def populate_manifest(self, image, media_type, return_headers, manifest):
        """Populate fake quay client with manifest for given media_type."""
        self.fake_manifests.setdefault(image, TDict[str, str]({}))
        self.fake_manifests[image][media_type] = manifest

    def get_manifest(
        self,
        image: str,
        raw: bool = False,
        media_type: str | None = None,
        return_headers: bool = False,
    ) -> ManifestList | Manifest | str:
        """
        Get manifest of given media type.

        Args:
            image (str):
                Image for which to get the manifest list.
            raw (bool):
                Whether to return the manifest as raw JSON.
            manifest_list (bool):
                Whether to only return a manifest list and raise an exception otherwise.
            media_type (str):
                Can be application/vnd.docker.distribution.manifest.list.v2+json,
                application/vnd.docker.distribution.manifest.v2+json,
                application/vnd.docker.distribution.manifest.v1+json,
                application/vnd.oci.image.manifest.v1+json, application/vnd.oci.image.index.v1+json
                or None indicating which manifest type is requested. If it's None,
                manifest list is prefered, but if v2s2 is returned instead, v2s2
                is returned as final result. If neither is found, same order is attempted with
                OCI type images.
        Returns (dict|str):
            Image manifest
        Raises:
            ManifestTypeError:
                When the image doesn't return the requested manifest type.
            ValueError:
                If Manifest list and V2S1 manifest are requested at the same time.
        """
        if media_type is not None:
            if image in self.fake_manifests and media_type in self.fake_manifests[image]:
                if raw:
                    if return_headers:
                        return (self.fake_manifests[image][media_type], {"return": "headers"})
                    else:
                        return self.fake_manifests[image][media_type]
                else:
                    return json.loads(self.fake_manifests[image][media_type])
            elif image in self.fake_manifests and media_type not in self.fake_manifests[image]:
                raise ManifestTypeError(
                    "Image {0} doesn't have a {1} manifest".format(image, media_type)
                )
            else:
                raise ManifestNotFoundError()
        else:
            for manifest_type in (
                QuayClient._MANIFEST_LIST_TYPE,
                QuayClient._MANIFEST_V2S2_TYPE,
                QuayClient._MANIFEST_OCI_LIST_TYPE,
                QuayClient._MANIFEST_OCI_V2S2_TYPE,
                QuayClient._MANIFEST_V2S1_TYPE,
            ):
                if image in self.fake_manifests and manifest_type in self.fake_manifests[image]:
                    if raw:
                        if return_headers:
                            return (
                                self.fake_manifests[image][manifest_type],
                                {"return": "headers"},
                            )
                        else:
                            return self.fake_manifests[image][manifest_type]
                    else:
                        return json.loads(self.fake_manifests[image][manifest_type])
            else:
                raise ManifestNotFoundError()

    def upload_manifest(self, manifest: ManifestList | str, image: str, raw: bool = False) -> None:
        """
        Upload manifest to a specified image.

        All manifest types are supported (manifest, manifest list).

        Args:
            manifest (dict):
                Manifest to be uploaded.
            image (str):
                Image address to upload the manifest to.
            raw (bool):
                Whether the given manifest is a string (raw) or a Python dictionary
        """
        if raw:
            manifest_type = json.loads(cast(str, manifest)).get(
                "mediaType", self._MANIFEST_V2S1_TYPE
            )
            self.fake_manifests.setdefault(image, TDict[str, str]({}))[manifest_type] = manifest
        else:
            manifest_type = cast(ManifestList, manifest).get("mediaType", self._MANIFEST_V2S1_TYPE)
            self.fake_manifests.setdefault(image, TDict[str, str]({}))[manifest_type] = json.dumps(
                manifest
            )



import json
import pytest

from pytractions.base import TDict

from signtractions.resources import fake_quay_client
from signtractions.resources import exceptions


def test_get_manifest_list_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        False,
        json.dumps(ml),
    )
    ret_ml = client.get_manifest(
        "quay.io/namespace/image:1",
        media_type="application/vnd.docker.distribution.manifest.list.v2+json",
    )
    assert ml == ret_ml


def test_get_manifest_list_raw_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        True,
        json.dumps(ml),
    )
    ret_ml = client.get_manifest(
        "quay.io/namespace/image:1",
        raw=True,
        media_type="application/vnd.docker.distribution.manifest.list.v2+json",
    )
    assert json.dumps(ml) == ret_ml


def test_get_manifest_list_raw_return_headers_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        {"some": "value"},
        json.dumps(ml, sort_keys=True),
    )
    ret_ml, headers = client.get_manifest(
        "quay.io/namespace/image:1",
        raw=True,
        return_headers=True,
        media_type="application/vnd.docker.distribution.manifest.list.v2+json",
    )
    assert json.dumps(ml, sort_keys=True) == ret_ml
    assert headers == {"return": "headers"}


def test_get_manifest_list_wrong_type():
    manifest = {
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
        "size": 429,
        "digest": "sha256:6d5f4d65fg4d6f54g",
        "platform": {"architecture": "arm64", "os": "linux"},
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(manifest),
    )

    with pytest.raises(exceptions.ManifestTypeError, match=".*doesn't have a .* manifest"):
        client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.list.v2+json",
        )


def test_get_manifest_list_not_found():
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    with pytest.raises(exceptions.ManifestNotFoundError):
        client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.list.v2+json",
        )


def test_get_manifest_accept_any():
    v2s1_manifest = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 1,
        "signatures": [],
    }

    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.v1+json",
        False,
        json.dumps(v2s1_manifest),
    )
    ret_manifest = client.get_manifest("quay.io/namespace/image:1")
    assert v2s1_manifest == ret_manifest


def test_get_manifest_raw_accept_any_return_headers():
    v2s1_manifest = {
        "name": "hello-world",
        "tag": "latest",
        "architecture": "amd64",
        "fsLayers": [],
        "history": [],
        "schemaVersion": 1,
        "signatures": [],
    }

    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.v1+json",
        False,
        json.dumps(v2s1_manifest),
    )
    ret_manifest, ret_headers = client.get_manifest(
        "quay.io/namespace/image:1", return_headers=True, raw=True
    )
    assert json.dumps(v2s1_manifest) == ret_manifest
    assert ret_headers == {"return": "headers"}


def test_get_v2s1_manifest_wrong_type():
    v2s2_manifest = {
        "mediaType": "application/vnd.docker.distribution.manifest.v1+json",
        "size": 429,
        "digest": "sha256:6d5f4d65fg4d6f54g",
        "platform": {"architecture": "arm64", "os": "linux"},
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.v1+json",
        False,
        json.dumps(v2s2_manifest, sort_keys=True),
    )
    with pytest.raises(
        exceptions.ManifestTypeError,
        match=r".*doesn't have a application/vnd\."
        r"docker\.distribution\.manifest\.v2\+json manifest",
    ):
        client.get_manifest(
            "quay.io/namespace/image:1",
            media_type="application/vnd.docker.distribution.manifest.v2+json",
        )


def test_upload_manifest_list_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.upload_manifest(json.dumps(ml, sort_keys=True), "quay.io/namespace/image:1", raw=True)
    assert client.fake_manifests["quay.io/namespace/image:1"][
        "application/vnd.docker.distribution.manifest.list.v2+json"
    ] == json.dumps(ml, sort_keys=True)


def test_upload_raw_manifest_success():
    ml = {
        "schemaVersion": 2,
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
        "manifests": [
            {
                "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
                "size": 429,
                "digest": "sha256:6d5f4d65fg4d6f54g",
                "platform": {"architecture": "arm64", "os": "linux"},
            }
        ],
    }
    client = fake_quay_client.FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    client.upload_manifest(json.dumps(ml), "quay.io/namespace/image:1", raw=True)
    assert client.fake_manifests["quay.io/namespace/image:1"][
        "application/vnd.docker.distribution.manifest.list.v2+json"
    ] == json.dumps(ml)



import contextlib
import functools
from dataclasses import dataclass, field
from io import StringIO
import logging
import os
import pkg_resources
import textwrap
from typing import Iterable, Any, Dict, Generator, Tuple, Optional, List, Callable
import sys

LOG = logging.getLogger("pubtools.quay")

INTERNAL_DELIMITER = "----"
MAX_RETRY_WAIT = 120
# from pubtools.pluggy import pm


@dataclass
class FData:
    """Dataclass for holding data for a function execution.

    Args:
        args (Iterable[Any]): Arguments for the function.
        kwargs (Dict[str, Any]): Keyword arguments for the function.
    """

    args: Iterable[Any]
    kwargs: Dict[str, Any] = field(default_factory=dict)


@contextlib.contextmanager
def capture_stdout() -> Generator[StringIO, None, None]:
    """Capture sys.stdout to stream buffer."""
    old_stdout = sys.stdout
    sys.stdout = new_stdout = StringIO()

    try:
        yield new_stdout
    finally:
        sys.stdout = old_stdout


@contextlib.contextmanager
def setup_entry_point_cli(
    entry_tuple: Tuple[str, str, str],
    name: Optional[str],
    args: List[str],
    environ_vars: Dict[str, Any],
) -> Generator[Callable[[], Any], None, None]:
    """
    Set up an entrypoint as a context manager.

    Args:
        entry_tuple ((str, str, str)):
            Tuple consisting of dependency, category, and entrypoint.
        name: (str):
            Entrypoint name.
        args ([str]):
            Entrypoint arguments.
        environ_vars (dict):
            Env variable names and values to set for the entrypoint.
    """
    orig_environ = os.environ.copy()

    try:
        # First argv element is always the entry point name.
        # For a console_scripts entry point, this will be the same value
        # as if the script was invoked directly. For any other kind of entry point,
        # this value is probably meaningless.
        for key in environ_vars:
            os.environ[key] = environ_vars[key]
        entry_point_func = pkg_resources.load_entry_point(*entry_tuple)
        if args:
            func_args = [name]
            func_args.extend(args)
            yield functools.partial(entry_point_func, func_args)
        else:
            yield entry_point_func
    finally:
        os.environ.update(orig_environ)

        to_delete = [key for key in os.environ if key not in orig_environ]
        for key in to_delete:
            del os.environ[key]


def run_entrypoint(
    entry_tuple: tuple[str, str, str], name: str, args: list[str], environ_vars: dict[str, Any]
) -> Any:
    """
    Run an entrypoint function and return its return value.

    Args:
        entry_tuple ((str, str, str)):
            Tuple consisting of dependency, category, and entrypoint.
        name: (str):
            Entrypoint name.
        args ([str]):
            Entrypoint arguments.
        environ_vars (dict):
            Env variable names and values to set for the entrypoint.

    Returns (str):
        Data returned by the entrypoint.
    """
    raw_args = " ".join([entry_tuple[2]] + args)
    wrapped_lines = textwrap.wrap(
        raw_args, 100, subsequent_indent="  ", break_on_hyphens=False, break_long_words=False
    )

    LOG.info("Running task with arguments:")
    for idx, line in enumerate(wrapped_lines):
        suffix = ""
        if idx != len(wrapped_lines) - 1:
            # shell-style backslash to indicate continuation
            suffix = " \\"
        LOG.info("%s%s", line, suffix)

    with setup_entry_point_cli(entry_tuple, name, args, environ_vars) as entry_func:
        with capture_stdout():
            pyret = entry_func()

    return pyret


def run_entrypoint_mod(entry_tuple: tuple[str, str, str], name: str, args: list[str]) -> Any:
    """Run entrypoint as python module function."""
    entry_point_func = pkg_resources.load_entry_point(*entry_tuple)
    pyret = entry_point_func(*args)

    return pyret



from typing import Optional
from dataclasses import field

from pytractions.base import Base, TList


class ContainerParts(Base):
    """Data structure to hold container reference parts."""

    registry: str = ""
    image: str = ""
    tag: Optional[str] = None
    digests: TList[str] = field(default_factory=TList[str])
    arches: TList[str] = field(default_factory=TList[str])

    def make_reference(self):
        """Convert container parts back to reference."""
        if self.tag:
            return "{}/{}:{}".format(self.registry, self.image, self.tag)
        else:
            return "{}/{}@{}".format(self.registry, self.image, self.digests[0])



import json
from typing import Union

import pytest
from pytractions.base import Res, Arg, In, STMDExecutorType, TDict, TList

from signtractions.tractors.t_sign_snapshot import SignSnapshot
from signtractions.resources.signing_wrapper import CosignSignerSettings
from signtractions.models.signing import SignEntry

from signtractions.resources.fake_signing_wrapper import FakeCosignSignerWrapper, FakeEPRunArgs
from signtractions.resources.fake_quay_client import FakeQuayClient


@pytest.fixture
def fake_cosign_wrapper():
    return FakeCosignSignerWrapper(
        config_file="test",
        settings=CosignSignerSettings(),
        fake_entry_point_requests=TList[FakeEPRunArgs]([]),
        fake_entry_point_returns=TList[TDict[str, TDict[str, str]]]([]),
        fake_entry_point_runs=TList[FakeEPRunArgs]([]),
    )


@pytest.fixture
def fake_quay_client():
    return FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )


def test_sign_snapshot(fix_manifest_v2s2, fix_snapshot_str, fake_cosign_wrapper, fake_quay_client):
    fake_quay_client.populate_manifest(
        "quay.io/containers/podman:latest",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(fix_manifest_v2s2),
    )
    fake_quay_client.populate_manifest(
        "quay.io/containers/podman:1.0",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(fix_manifest_v2s2),
    )
    fake_cosign_wrapper.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": [
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                    ],
                    "reference": [
                        "quay.io/containers/podman:latest",
                        "quay.io/containers/podman:1.0",
                    ],
                    "signing_key": "signing_key",
                }
            ),
        )
    )
    fake_cosign_wrapper.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]].content_from_json({"signer_result": {"status": "ok"}})
    )
    t = SignSnapshot(
        uid="test",
        a_pool_size=Arg[int](a=1),
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.LOCAL),
        r_dst_quay_client=Res[FakeQuayClient](r=fake_quay_client),
        r_signer_wrapper_cosign=Res[FakeCosignSignerWrapper](r=fake_cosign_wrapper),
        i_task_id=In[int](data=1),
        i_snapshot_str=In[str](data=fix_snapshot_str),
        i_signing_key=In[str](data="signing_key"),
    )
    t.run()
    assert len(t.o_sign_entries.data) == 2
    assert t.o_sign_entries.data[0] == SignEntry(
        reference="quay.io/containers/podman:latest",
        repo="containers/podman",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )
    assert t.o_sign_entries.data[0] == SignEntry(
        reference="quay.io/containers/podman:latest",
        repo="containers/podman",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )


def test_sign_snapshot_file(
    fix_manifest_v2s2, fix_snapshot_file, fake_cosign_wrapper, fake_quay_client
):
    fake_quay_client.populate_manifest(
        "quay.io/containers/podman:latest",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(fix_manifest_v2s2),
    )
    fake_quay_client.populate_manifest(
        "quay.io/containers/podman:1.0",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(fix_manifest_v2s2),
    )
    fake_cosign_wrapper.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": [
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
                    ],
                    "reference": [
                        "quay.io/containers/podman:latest",
                        "quay.io/containers/podman:1.0",
                    ],
                    "signing_key": "signing_key",
                }
            ),
        )
    )
    fake_cosign_wrapper.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]].content_from_json({"signer_result": {"status": "ok"}})
    )

    t = SignSnapshot(
        uid="test",
        a_pool_size=Arg[int](a=1),
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.LOCAL),
        r_dst_quay_client=Res[FakeQuayClient](r=fake_quay_client),
        r_signer_wrapper_cosign=Res[FakeCosignSignerWrapper](r=fake_cosign_wrapper),
        i_task_id=In[int](data=1),
        i_snapshot_file=In[str](data=fix_snapshot_file),
        i_signing_key=In[str](data="signing_key"),
    )
    t.run()
    assert len(t.o_sign_entries.data) == 2
    assert t.o_sign_entries.data[0] == SignEntry(
        reference="quay.io/containers/podman:latest",
        repo="containers/podman",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )
    assert t.o_sign_entries.data[0] == SignEntry(
        reference="quay.io/containers/podman:latest",
        repo="containers/podman",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )



import json
from typing import Union

import pytest
from pytractions.base import TList, TDict, Res, Arg, In

from signtractions.tractors.t_sign_containers import SignContainers
from signtractions.resources.signing_wrapper import CosignSignerSettings
from signtractions.models.containers import ContainerParts
from signtractions.models.signing import SignEntry

from signtractions.resources.fake_signing_wrapper import FakeCosignSignerWrapper, FakeEPRunArgs
from signtractions.resources.fake_quay_client import FakeQuayClient


@pytest.fixture
def fake_cosign_wrapper():
    return FakeCosignSignerWrapper(
        config_file="test",
        settings=CosignSignerSettings(),
        fake_entry_point_requests=TList[FakeEPRunArgs]([]),
        fake_entry_point_returns=TList[TDict[str, TDict[str, str]]]([]),
        fake_entry_point_runs=TList[FakeEPRunArgs]([]),
    )


@pytest.fixture
def fake_quay_client():
    return FakeQuayClient(
        username="user",
        password="pass",
        host="quay.io",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )


def test_sign_containers_tags(fix_manifest_v2s2, fake_cosign_wrapper, fake_quay_client):
    fake_quay_client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(fix_manifest_v2s2),
    )
    fake_cosign_wrapper.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": TList[str](
                        ["sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d"]
                    ),
                    "reference": TList[str](["quay.io/namespace/image:1"]),
                    "signing_key": "signing_key",
                },
            ),
        )
    )
    fake_cosign_wrapper.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]].content_from_json({"signer_result": {"status": "ok"}})
    )

    t = SignContainers(
        uid="test",
        r_signer_wrapper_cosign=Res[FakeCosignSignerWrapper](r=fake_cosign_wrapper),
        a_pool_size=Arg[int](a=1),
        r_dst_quay_client=Res[FakeQuayClient](r=fake_quay_client),
        i_task_id=In[int](data=1),
        i_signing_keys=In[TList[str]](data=TList[str](["signing_key"])),
        i_container_image_references=In[TList[str]](data=TList[str](["quay.io/namespace/image:1"])),
    )
    t.run()
    assert len(t.tractions["t_parse_container_references"].o_container_parts.data) == 1
    assert t.tractions["t_parse_container_references"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag="1",
        digests=TList[str](),
        arches=TList[str](),
    )
    assert len(t.tractions["t_populate_digests"].o_container_parts.data) == 1
    assert t.tractions["t_populate_digests"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag="1",
        digests=TList[str](
            ["sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d"]
        ),
        arches=TList[str]([""]),
    )
    assert t.tractions["t_sign_entries_from_push_item"].o_sign_entries.data[0][0] == SignEntry(
        reference="quay.io/namespace/image:1",
        repo="namespace/image",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )
    assert t.o_sign_entries.data[0] == SignEntry(
        reference="quay.io/namespace/image:1",
        repo="namespace/image",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )


def test_sign_containers_tags_ml(fix_manifest_list, fake_cosign_wrapper, fake_quay_client):
    fake_quay_client.populate_manifest(
        "quay.io/namespace/image:1",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        False,
        json.dumps(fix_manifest_list),
    )
    fake_cosign_wrapper.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": [
                        "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
                        "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
                        "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
                        "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
                        "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
                        "sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3",
                    ],
                    "reference": [
                        "quay.io/namespace/image:1",
                        "quay.io/namespace/image:1",
                        "quay.io/namespace/image:1",
                        "quay.io/namespace/image:1",
                        "quay.io/namespace/image:1",
                        "quay.io/namespace/image:1",
                    ],
                    "signing_key": "signing_key",
                }
            ),
        )
    )
    fake_cosign_wrapper.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]].content_from_json({"signer_result": {"status": "ok"}})
    )

    t = SignContainers(
        uid="test",
        r_signer_wrapper_cosign=Res[FakeCosignSignerWrapper](r=fake_cosign_wrapper),
        r_dst_quay_client=Res[FakeQuayClient](r=fake_quay_client),
        i_task_id=In[int](data=1),
        a_pool_size=Arg[int](a=1),
        i_signing_keys=In[TList[str]](data=TList[str](["signing_key"])),
        i_container_image_references=In[TList[str]](data=TList[str](["quay.io/namespace/image:1"])),
    )
    t.run()
    assert len(t.tractions["t_parse_container_references"].o_container_parts.data) == 1
    assert t.tractions["t_parse_container_references"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag="1",
        digests=TList[str](),
        arches=TList[str](),
    )
    assert len(t.tractions["t_populate_digests"].o_container_parts.data) == 1
    assert t.tractions["t_populate_digests"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag="1",
        digests=TList[str](
            [
                "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
                "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
                "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
                "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
                "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
                "sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3",
            ]
        ),
        arches=TList[str](["amd64", "arm64", "arm", "ppc64le", "s390x", "multiarch"]),
    )
    assert len(t.tractions["t_sign_entries_from_push_item"].o_sign_entries.data[0]) == 6
    assert t.tractions["t_sign_entries_from_push_item"].o_sign_entries.data[0] == TList[SignEntry](
        [
            SignEntry(
                reference="quay.io/namespace/image:1",
                repo="namespace/image",
                digest="sha256:2e8f38a0a8d2a450598430fa70c7f0b53a" "eec991e76c3e29c63add599b4ef7ee",
                arch="amd64",
                signing_key="signing_key",
            ),
            SignEntry(
                reference="quay.io/namespace/image:1",
                repo="namespace/image",
                digest="sha256:b3f9218fb5839763e62e52ee6567fe331a" "a1f3c644f9b6f232ff23959257acf9",
                arch="arm64",
                signing_key="signing_key",
            ),
            SignEntry(
                reference="quay.io/namespace/image:1",
                repo="namespace/image",
                digest="sha256:496fb0ff2057c79254c9dc6ba999608a982" "19c5c93142569a547277c679e532c",
                arch="arm",
                signing_key="signing_key",
            ),
            SignEntry(
                reference="quay.io/namespace/image:1",
                repo="namespace/image",
                digest="sha256:146ab6fa7ba3ab4d154b09c1c5522e4966e" "cd071bf23d1ba3df6c8b9fc33f8cb",
                arch="ppc64le",
                signing_key="signing_key",
            ),
            SignEntry(
                reference="quay.io/namespace/image:1",
                repo="namespace/image",
                digest="sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09d" "ab7ffe64be1ae3ae66e76275eabd",
                arch="s390x",
                signing_key="signing_key",
            ),
            SignEntry(
                reference="quay.io/namespace/image:1",
                repo="namespace/image",
                digest="sha256:d07476154b88059d730e260eba282b3c7a0b" "5e7feb620638d49070b71dcdcaf3",
                arch="multiarch",
                signing_key="signing_key",
            ),
        ]
    )


def test_sign_containers_digests(fix_manifest_v2s2, fake_cosign_wrapper, fake_quay_client):
    fake_quay_client.populate_manifest(
        "quay.io/namespace/image@sha256:6ef06d8c90c863ba4eb4297f10"
        "73ba8cb28c1f6570e2206cdaad2084e2a4715d",
        "application/vnd.docker.distribution.manifest.v2+json",
        False,
        json.dumps(fix_manifest_v2s2),
    )
    fake_cosign_wrapper.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": [
                        "sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d"
                    ],
                    "reference": [None],
                    "signing_key": "signing_key",
                }
            ),
        )
    )
    fake_cosign_wrapper.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]].content_from_json({"signer_result": {"status": "ok"}})
    )

    t = SignContainers(
        uid="test",
        r_signer_wrapper_cosign=Res[FakeCosignSignerWrapper](r=fake_cosign_wrapper),
        a_pool_size=Arg[int](a=1),
        r_dst_quay_client=Res[FakeQuayClient](r=fake_quay_client),
        i_task_id=In[int](data=1),
        i_signing_keys=In[TList[str]](data=TList[str](["signing_key"])),
        i_container_image_references=In[TList[str]](
            data=TList[str](
                [
                    "quay.io/namespace/image@sha256:6ef06d8c90c863ba4eb4297f1073ba8c"
                    "b28c1f6570e2206cdaad2084e2a4715d"
                ]
            )
        ),
    )
    t.run()
    assert len(t.tractions["t_parse_container_references"].o_container_parts.data) == 1
    assert t.tractions["t_parse_container_references"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag=None,
        digests=TList[str](
            ["sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d"]
        ),
        arches=TList[str]([""]),
    )
    assert len(t.tractions["t_populate_digests"].o_container_parts.data) == 1
    assert t.tractions["t_populate_digests"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag=None,
        digests=TList[str](
            ["sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d"]
        ),
        arches=TList[str]([""]),
    )
    assert t.tractions["t_sign_entries_from_push_item"].o_sign_entries.data[0][0] == SignEntry(
        reference=None,
        repo="namespace/image",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )
    assert t.o_sign_entries.data[0] == SignEntry(
        reference=None,
        repo="namespace/image",
        digest="sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d",
        arch="",
        signing_key="signing_key",
    )


def test_sign_containers_digests_ml(fix_manifest_list, fake_cosign_wrapper, fake_quay_client):
    fake_quay_client.populate_manifest(
        "quay.io/namespace/image@sha256:d07476154b88059d730e260eba282b3c"
        "7a0b5e7feb620638d49070b71dcdcaf3",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        False,
        json.dumps(fix_manifest_list),
    )
    fake_cosign_wrapper.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": [
                        "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
                        "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
                        "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
                        "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
                        "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
                        "sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3",
                    ],
                    "reference": [None, None, None, None, None, None],
                    "signing_key": "signing_key",
                }
            ),
        )
    )
    fake_cosign_wrapper.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]]({"signer_result": TDict[str, str]({"status": "ok"})})
    )

    t = SignContainers(
        uid="test",
        r_signer_wrapper_cosign=Res[FakeCosignSignerWrapper](r=fake_cosign_wrapper),
        r_dst_quay_client=Res[FakeQuayClient](r=fake_quay_client),
        i_task_id=In[int](data=1),
        a_pool_size=Arg[int](a=1),
        i_signing_keys=In[TList[str]](data=TList[str](["signing_key"])),
        i_container_image_references=In[TList[str]](
            data=TList[str](
                [
                    "quay.io/namespace/image@sha256:d07476154b88059d730e260eba"
                    "282b3c7a0b5e7feb620638d49070b71dcdcaf3"
                ]
            )
        ),
    )
    t.run()
    assert len(t.tractions["t_parse_container_references"].o_container_parts.data) == 1
    assert t.tractions["t_parse_container_references"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag=None,
        digests=TList[str](
            ["sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3"]
        ),
        arches=TList[str]([""]),
    )
    assert len(t.tractions["t_populate_digests"].o_container_parts.data) == 1
    assert t.tractions["t_populate_digests"].o_container_parts.data[0] == ContainerParts(
        registry="quay.io",
        image="namespace/image",
        tag=None,
        digests=TList[str](
            [
                "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
                "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
                "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
                "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
                "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
                "sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3",
            ]
        ),
        arches=TList[str](["amd64", "arm64", "arm", "ppc64le", "s390x", "multiarch"]),
    )
    assert len(t.tractions["t_sign_entries_from_push_item"].o_sign_entries.data[0]) == 6
    assert t.tractions["t_sign_entries_from_push_item"].o_sign_entries.data[0] == TList[SignEntry](
        [
            SignEntry(
                reference=None,
                repo="namespace/image",
                digest="sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec99" "1e76c3e29c63add599b4ef7ee",
                arch="amd64",
                signing_key="signing_key",
            ),
            SignEntry(
                reference=None,
                repo="namespace/image",
                digest="sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644" "f9b6f232ff23959257acf9",
                arch="arm64",
                signing_key="signing_key",
            ),
            SignEntry(
                reference=None,
                repo="namespace/image",
                digest="sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93" "142569a547277c679e532c",
                arch="arm",
                signing_key="signing_key",
            ),
            SignEntry(
                reference=None,
                repo="namespace/image",
                digest="sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf" "23d1ba3df6c8b9fc33f8cb",
                arch="ppc64le",
                signing_key="signing_key",
            ),
            SignEntry(
                reference=None,
                repo="namespace/image",
                digest="sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe6" "4be1ae3ae66e76275eabd",
                arch="s390x",
                signing_key="signing_key",
            ),
            SignEntry(
                reference=None,
                repo="namespace/image",
                digest="sha256:d07476154b88059d730e260eba282" "b3c7a0b5e7feb620638d49070b71dcdcaf3",
                arch="multiarch",
                signing_key="signing_key",
            ),
        ]
    )



from unittest import mock
import pytest
import logging

from signtractions.resources import command_executor
from .utils.misc import compare_logs


def test_local_executor_init():
    executor = command_executor.LocalExecutor({"some_param": "value"})

    assert executor.params["some_param"] == "value"
    assert executor.params["universal_newlines"] is True
    assert executor.params["stderr"] == -1
    assert executor.params["stdout"] == -1
    assert executor.params["stdin"] == -1


@mock.patch("signtractions.resources.command_executor.subprocess.Popen")
def test_local_executor_run(mock_popen):
    executor = command_executor.LocalExecutor({"some_param": "value"})

    mock_communicate = mock.MagicMock()
    mock_communicate.return_value = ("outlog", "errlog")
    mock_popen.return_value.communicate = mock_communicate
    mock_popen.return_value.returncode = 0

    out, err = executor._run_cmd("pwd", stdin="input")
    assert out == "outlog"
    assert err == "errlog"
    mock_popen.assert_called_once_with(
        ["pwd"],
        some_param="value",
        universal_newlines=True,
        stderr=-1,
        stdout=-1,
        stdin=-1,
    )
    mock_communicate.assert_called_once_with(input="input")


@mock.patch("signtractions.resources.command_executor.subprocess.Popen")
def test_local_executor_context_manager(mock_popen):
    mock_communicate = mock.MagicMock()
    mock_communicate.return_value = ("outlog", "errlog")
    mock_popen.return_value.communicate = mock_communicate
    mock_popen.return_value.returncode = 0

    with command_executor.LocalExecutor({"some_param": "value"}) as executor:
        out, err = executor._run_cmd("pwd", stdin="input")
    assert out == "outlog"
    assert err == "errlog"
    mock_popen.assert_called_once_with(
        ["pwd"],
        some_param="value",
        universal_newlines=True,
        stderr=-1,
        stdout=-1,
        stdin=-1,
    )
    mock_communicate.assert_called_once_with(input="input")


@mock.patch("signtractions.resources.command_executor.subprocess.Popen")
def test_local_executor_run_error(mock_popen):
    executor = command_executor.LocalExecutor({"some_param": "value"})

    mock_communicate = mock.MagicMock()
    mock_communicate.return_value = ("outlog", "errlog")
    mock_popen.return_value.communicate = mock_communicate
    mock_popen.return_value.returncode = -1

    with pytest.raises(RuntimeError, match="An error has occured when executing.*"):
        executor._run_cmd("pwd", stdin="input")


@mock.patch("signtractions.resources.command_executor.subprocess.Popen")
def test_local_executor_run_long_error(mock_popen, caplog):
    caplog.set_level(logging.ERROR)
    executor = command_executor.LocalExecutor({"some_param": "value"})

    err_msg = " ".join(["Very long error message."] * 40)

    mock_communicate = mock.MagicMock()
    mock_communicate.return_value = ("outlog", err_msg)
    mock_popen.return_value.communicate = mock_communicate
    mock_popen.return_value.returncode = -1

    expected_logs = [
        ".*failed with the following error:",
        "    Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message.",
        "    Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message.",
        "    Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message.",
        "    Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message.",
        "    Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message. Very long error message. "
        "Very long error message. Very long error message.",
    ]

    with pytest.raises(RuntimeError, match="An error has occured when executing.*"):
        executor._run_cmd("pwd", stdin="input")

    compare_logs(caplog, expected_logs)


@mock.patch("signtractions.resources.command_executor.subprocess.Popen")
def test_local_executor_run_error_custom_message(mock_popen):
    executor = command_executor.LocalExecutor({"some_param": "value"})

    mock_communicate = mock.MagicMock()
    mock_communicate.return_value = ("outlog", "errlog")
    mock_popen.return_value.communicate = mock_communicate
    mock_popen.return_value.returncode = -1

    with pytest.raises(RuntimeError, match="Custom error"):
        executor._run_cmd("pwd", stdin="input", err_msg="Custom error")


@mock.patch("signtractions.resources.command_executor.subprocess.Popen")
def test_local_executor_run_tolerate_err(mock_popen):
    executor = command_executor.LocalExecutor({"some_param": "value"})

    mock_communicate = mock.MagicMock()
    mock_communicate.return_value = ("outlog", "errlog")
    mock_popen.return_value.communicate = mock_communicate
    mock_popen.return_value.returncode = -1

    out, err = executor._run_cmd("pwd", stdin="input", tolerate_err=True)
    assert out == "outlog"
    assert err == "errlog"
    mock_popen.assert_called_once_with(
        ["pwd"],
        some_param="value",
        universal_newlines=True,
        stderr=-1,
        stdout=-1,
        stdin=-1,
    )
    mock_communicate.assert_called_once_with(input="input")


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_login_already_logged(mock_run_cmd):
    executor = command_executor.LocalExecutor()

    mock_run_cmd.return_value = ("quay_user", "")
    executor.skopeo_login("quay_host", "quay_user", "quay_token")
    mock_run_cmd.assert_called_once_with("skopeo login --get-login quay_host", tolerate_err=True)


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_login_missing_credentials(mock_run_cmd):
    executor = command_executor.LocalExecutor()

    mock_run_cmd.return_value = ("not logged into quay", "")
    with pytest.raises(ValueError, match=".*login credentials are not present.*"):
        executor.skopeo_login("some-host")


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_login_success(mock_run_cmd):
    executor = command_executor.LocalExecutor()

    mock_run_cmd.side_effect = [("not logged into quay", ""), ("Login Succeeded", "")]
    executor.skopeo_login("quay_host", "quay_user", "quay_token")
    assert mock_run_cmd.call_args_list == [
        mock.call("skopeo login --get-login quay_host", tolerate_err=True),
        mock.call(
            "skopeo login -u quay_user --password-stdin quay_host",
            stdin="quay_token",
        ),
    ]


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_login_failed(mock_run_cmd):
    executor = command_executor.LocalExecutor()

    mock_run_cmd.side_effect = [("not logged into quay", ""), ("", "Login failed")]
    with pytest.raises(RuntimeError, match="Login command didn't generate.*"):
        executor.skopeo_login("quay_host", "quay_user", "quay_token")


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_tag_images(mock_run_cmd):
    executor = command_executor.LocalExecutor()

    executor.tag_images("quay.io/repo/image:1", ["quay.io/repo/dest:1", "quay.io/repo/dest:2"])
    assert mock_run_cmd.call_args_list == [
        mock.call(
            "skopeo copy docker://quay.io/repo/image:1 docker://quay.io/repo/dest:1 --format v2s2"
        ),
        mock.call(
            "skopeo copy docker://quay.io/repo/image:1 docker://quay.io/repo/dest:2 --format v2s2"
        ),
    ]


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_tag_images_all_arch(mock_run_cmd):
    executor = command_executor.LocalExecutor()

    executor.tag_images(
        "quay.io/repo/image:1", ["quay.io/repo/dest:1", "quay.io/repo/dest:2"], True
    )
    assert mock_run_cmd.call_args_list == [
        mock.call(
            "skopeo copy --all docker://quay.io/repo/image:1 docker://quay.io/repo/dest:1"
            " --format v2s2"
        ),
        mock.call(
            "skopeo copy --all docker://quay.io/repo/image:1 docker://quay.io/repo/dest:2"
            " --format v2s2"
        ),
    ]


@mock.patch("signtractions.resources.command_executor.LocalExecutor._run_cmd")
def test_skopeo_inspect(mock_run_cmd):
    mock_run_cmd.return_value = ('{"aaa":"bbb"}', "")
    executor = command_executor.LocalExecutor()

    ret = executor.skopeo_inspect("quay.io/repo/image:1")
    mock_run_cmd.assert_called_once_with("skopeo inspect docker://quay.io/repo/image:1")
    assert ret == {"aaa": "bbb"}

    ret = executor.skopeo_inspect("quay.io/repo/image:1", raw=True)
    assert ret == '{"aaa":"bbb"}'



import json
import logging
from unittest import mock
import pytest
import requests_mock
import requests
import os

from signtractions.resources import exceptions
from signtractions.resources.utils import misc
from .utils.misc import sort_dictionary_sortable_values, compare_logs, mock_entry_point

# flake8: noqa: E501

LOG = logging.getLogger()
logging.basicConfig()
LOG.setLevel(logging.INFO)


def test_run_entrypoint(caplog):
    def side_effect_entrypoint(*args):
        LOG.warning("log generated by entrypoint")
        LOG.info("environment variable: {0}".format(os.environ["PYXIS_PASSWORD"]))
        return {"pyxis": "data"}

    with mock_entry_point(
        "pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"
    ) as entrypoint:
        entrypoint.side_effect = side_effect_entrypoint
        ret_val = misc.run_entrypoint(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"),
            "pubtools-pyxis-get-operator-indices",
            [
                "--pyxis-krb-principal",
                "some-principal@REDHAT.COM",
                "--organization",
                "redhat-operators",
                "--ocp-versions-range",
                "v4.5",
                "--pyxis-krb-ktfile",
                "/etc/pub/some.keytab",
            ],
            {"PYXIS_PASSWORD": "some-password"},
        )
        expected_logs = [
            "Running task with arguments:",
            "pubtools-pyxis-get-operator-indices --pyxis-krb-principal some-principal@REDHAT\.COM --organization \\\\",
            "redhat-operators --ocp-versions-range v4.5 --pyxis-krb-ktfile /etc/pub/some\.keytab",
            "log generated by entrypoint",
            "environment variable: some-password",
        ]
        compare_logs(caplog, expected_logs)


def test_run_entrypoint_mod(caplog):
    def side_effect_entrypoint(*args):
        LOG.warning("log generated by entrypoint")
        return {"return": "value"}

    with mock_entry_point(
        "pubtools-sign", "modules", "pubtools-sign-cosign-signature-list"
    ) as entrypoint:
        entrypoint.side_effect = side_effect_entrypoint
        ret_val = misc.run_entrypoint_mod(
            ("pubtools-sign", "modules", "pubtools-sign-cosign-signature-list"),
            None,
            ["some-ep-argument"],
        )
        expected_logs = [
            "log generated by entrypoint",
        ]
        compare_logs(caplog, expected_logs)


def test_run_entrypoint_without_args(caplog):
    def side_effect_entrypoint(*args):
        LOG.warning("log generated by entrypoint")
        LOG.info("environment variable: {0}".format(os.environ["PYXIS_PASSWORD"]))
        return {"pyxis": "data"}

    with mock_entry_point(
        "pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"
    ) as entrypoint:
        entrypoint.side_effect = side_effect_entrypoint
        ret_val = misc.run_entrypoint(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"),
            "pubtools-pyxis-get-operator-indices",
            [],
            {"PYXIS_PASSWORD": "some-password"},
        )
        expected_logs = [
            "Running task with arguments:",
            "pubtools-pyxis-get-operator-indices",
            "log generated by entrypoint",
            "environment variable: some-password",
        ]
        compare_logs(caplog, expected_logs)



from typing import Optional

from pytractions.base import Base


class SignEntry(Base):
    """Data structure to hold signing related information.

    Args:
        signing_key (str): Signing key.
        repo (str): Repo reference in format <registry>/<repo>
        reference (str): Reference in format <registry>/<repo>:<tag>
        digest (str): Digest of the manifest.
        arch (str): Architecture of the manifest.
    """

    repo: str
    reference: Optional[str]
    digest: str
    signing_key: str
    arch: str



import abc
from contextlib import contextmanager
import dataclasses
import logging
import pkg_resources
import tempfile
import json

from typing import Optional, List, Dict, Any, Tuple, Generator

from marshmallow import Schema

from pytractions.base import Base, doc

from .utils.misc import (
    run_entrypoint,
    run_entrypoint_mod,
)
from ..models.signing import SignEntry


LOG = logging.getLogger("pubtools.quay")


class SigningError(Exception):
    """Error raised when signing fails."""

    pass


class NoSchema(Schema):
    """Schema that does not validate anything."""

    pass


class SignerWrapperSettings(Base):
    """Signer wrapper settings."""

    pass


class SignerWrapper(Base):
    """Wrapper providing functionality to sign containers with a generic signer."""

    label: str = dataclasses.field(default="unused", init=False)
    pre_push: bool = dataclasses.field(default=False, init=False)
    _entry_point_conf = ["signer", "group", "signer"]
    config_file: Optional[str]
    settings: SignerWrapperSettings
    _ep: Optional[Any] = None

    d_config_file: str = doc("Path to pubtools-sign config file.")

    @property
    def entry_point(self) -> Any:
        """Load and return entry point for pubtools-sign project."""
        if self._ep is None:
            self._ep = pkg_resources.load_entry_point(*self._entry_point_conf)
        return self._ep

    def remove_signatures(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> None:
        """Remove signatures from a sigstore."""
        LOG.debug("Removing signatures %s", signatures)
        self._remove_signatures(signatures)

    @abc.abstractmethod
    def _run_remove_signatures(self, signatures_to_remove: List[Any]) -> None:
        pass  # pragma: no cover

    def _remove_signatures(self, signatures_to_remove: List[Any]) -> None:
        """Remove signatures from sigstore.

        This is helper to make testing easier.
        Args:
            signatures_to_remove (list): Signatures to remove.
        """
        self._run_remove_signatures(signatures_to_remove)

    @abc.abstractmethod
    def _run_store_signed(self, signatures: Dict[str, Any]) -> None:
        pass  # pragma: no cover

    def _store_signed(self, signatures: Dict[str, Any]) -> None:
        """Store signatures in sigstore.

        This is helper to make testing easier.
        Args:
            signatures (dict): Signatures to store.
        """
        LOG.debug("Storing signatures %s", signatures)
        self._run_store_signed(signatures)

    def sign_container_opt_args(
        self, sign_entry: SignEntry, task_id: Optional[str] = None
    ) -> Dict[str, Any]:
        """Return optional arguments for signing a container.

        Args:
            sign_entry (SignEntry): SignEntry to sign.
            task_id (str): Task ID to identify the signing task if needed.

        Returns:
            dict: Optional arguments for signing a container.
        """
        return {}

    @abc.abstractmethod
    def _filter_to_sign(self, to_sign_entries: List[SignEntry]) -> List[SignEntry]:
        """Filter entries to sign.

        Args:
            to_sign_entries (List[SignEntry]): list of entries to sign.

        Returns:
            List[SignEntry]: list of entries to sign.
        """
        return to_sign_entries  # pragma: no cover

    def sign_containers(
        self,
        sign_entries: List[SignEntry],
        task_id: Optional[str] = None,
    ) -> None:
        """Sign a specific chunk of references and digests with given signing key.

        Args:
            sign_entries (List[SignEntry]): Chunk of SignEntry to sign.
            task_id (str): Task ID to identify the signing task if needed.
        """
        for sign_entry in sign_entries:
            LOG.info(
                "Signing container %s %s %s",
                sign_entry.reference,
                sign_entry.digest,
                sign_entry.signing_key,
            )
        if not sign_entries:
            return
        sign_entry = sign_entries[0]
        opt_args = self.sign_container_opt_args(sign_entry, task_id)
        signed = self.entry_point(
            config_file=self.config_file,
            signing_key=sign_entry.signing_key,
            reference=[x.reference for x in sign_entries if x],
            digest=[x.digest for x in sign_entries if x],
            **opt_args,
        )
        if signed["signer_result"]["status"] != "ok":
            raise SigningError(signed["signer_result"]["error_message"])
        for sign_entry in sign_entries:
            LOG.info(
                "Signed %s(%s) with %s in %s",
                sign_entry.reference,
                sign_entry.digest,
                sign_entry.signing_key,
                self.label,
            )


class MsgSignerSettings(Base):
    """Validation schema for messaging signer settings."""

    pyxis_server: str
    pyxis_ssl_crt_file: Optional[str]
    pyxis_ssl_key_file: Optional[str]
    num_thread_pyxis: int = 7

    d_pyxis_server: str = doc("Pyxis server URL.")
    d_pyxis_ssl_crt_file: str = doc("Pyxis SSL client certificate file.")
    d_pyxis_ssl_key_file: str = doc("Pyxis SSL client key file.")
    d_num_thread_pyxis: str = doc("Number of threads to use for Pyxis requests.")


class MsgSignerWrapper(SignerWrapper):
    """Wrapper for messaging signer functionality."""

    label: str = dataclasses.field(default="msg_signer", init=False)
    pre_push: bool = dataclasses.field(default=True, init=False)
    settings: MsgSignerSettings

    _entry_point_conf = ["pubtools-sign", "modules", "pubtools-sign-msg-container-sign"]
    MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST: int = 50

    def _filter_to_sign(self, to_sign_entries: List[SignEntry]) -> List[SignEntry]:
        to_sign_digests = [x.digest for x in to_sign_entries]
        existing_signatures = [esig for esig in self._fetch_signatures(to_sign_digests)]
        existing_signatures_drk = {
            (x["manifest_digest"], x["reference"], x["sig_key_id"]) for x in existing_signatures
        }
        ret = []
        for tse in to_sign_entries:
            if (tse.digest, tse.reference, tse.signing_key) not in existing_signatures_drk:
                ret.append(tse)
        return ret

    def sign_container_opt_args(
        self, sign_entry: SignEntry, task_id: Optional[str] = None
    ) -> Dict[str, Any]:
        """Return optional arguments for signing a container.

        Args:
            sign_entry (SignEntry): SignEntry to sign.
            task_id (str): Task ID to identify the signing task if needed.

        Returns:
            dict: Optional arguments for signing a container.
        """
        return {k: v for k, v in [("task_id", task_id)] if v is not None}

    @contextmanager
    def _save_signatures_file(self, signatures: List[Dict[str, Any]]) -> Generator[Any, None, None]:
        """Save signatures to a temporary file and yield the file."""
        with tempfile.NamedTemporaryFile(
            mode="w", prefix="pubtools_quay_upload_signatures_"
        ) as signature_file:
            json.dump(signatures, signature_file)
            signature_file.flush()
            yield signature_file

    def _fetch_signatures(
        self, manifest_digests: List[str]
    ) -> Generator[dict[str, Any], None, None]:
        """Fetch signatures from sigstore.

        Args:
            manifest_digests (list): Manifest digests to fetch signatures for.
        Returns:
            List[Dict[str, Any]]: List of fetched signatures.
        """
        cert, key = self.settings.pyxis_ssl_crt_file, self.settings.pyxis_ssl_key_file
        chunk_size = self.MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST
        manifest_digests = [x for x in sorted(list(set(manifest_digests))) if x]

        args = ["--pyxis-server", self.settings.pyxis_server]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--request-threads", str(self.settings.num_thread_pyxis or 7)]

        for chunk_start in range(0, len(manifest_digests), chunk_size):
            chunk = manifest_digests[chunk_start : chunk_start + chunk_size]  # noqa: E203

            args = ["--pyxis-server", self.settings.pyxis_server]
            args += ["--pyxis-ssl-crtfile", cert]
            args += ["--pyxis-ssl-keyfile", key]

            with tempfile.NamedTemporaryFile(
                mode="w", prefix="pubtools_quay_get_signatures_"
            ) as signature_fetch_file:
                if manifest_digests:
                    json.dump(chunk, signature_fetch_file)
                    signature_fetch_file.flush()
                    args += ["--manifest-digest", "@{0}".format(signature_fetch_file.name)]

                env_vars: dict[Any, Any] = {}
                chunk_results = run_entrypoint(
                    ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-signatures"),
                    "pubtools-pyxis-get-signatures",
                    args,
                    env_vars,
                )

            for result in chunk_results:
                yield result

    def _run_store_signed(self, signed_results: Dict[str, Any]) -> None:
        """
        Upload signatures to Pyxis by using a pubtools-pyxis entrypoint.

        Data required for a Pyxis POST request:
        - manifest_digest
        - reference
        - repository
        - sig_key_id
        - signature_data

        Signatures are uploaded in batches.

        Args:
            signed_results: (Dict[str, Any]):
                Dictionary of {"signer_result":..., "operation_results":..., "signing_key":...}"}
                holding signed manifest claims data
        """
        LOG.info("Sending new signatures to Pyxis")

        signatures: List[Dict[str, Any]] = []
        for reference, op_res in zip(
            signed_results["operation"]["references"], signed_results["operation_results"]
        ):
            signatures.append(
                {
                    "manifest_digest": op_res[0]["msg"]["manifest_digest"],
                    "reference": reference,
                    "repository": op_res[0]["msg"]["repo"],
                    "sig_key_id": signed_results["signing_key"],
                    "signature_data": op_res[0]["msg"]["signed_claim"],
                }
            )

        for sig in signatures:
            LOG.debug(
                f"Uploading new signature. Reference: {sig['reference']}, "
                f"Repository: {sig['repository']}, "
                f"Digest: {sig['manifest_digest']}, "
                f"Key: {sig['sig_key_id']}"
            )

        cert, key = self.settings.pyxis_ssl_crt_file, self.settings.pyxis_ssl_key_file

        args = ["--pyxis-server", self.settings.pyxis_server]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--request-threads", str(self.settings.num_thread_pyxis or 7)]

        LOG.info("upload signature ARGS: %s", args)
        with self._save_signatures_file(signatures) as signature_file:
            LOG.info("upload signaturures %s", signatures)

            args += ["--signatures", "@{0}".format(signature_file.name)]
            LOG.info("Uploading {0} new signatures".format(len(signatures)))
            env_vars: dict[Any, Any] = {}
            run_entrypoint_mod(
                ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-upload-signatures"),
                "pubtools-pyxis-upload-signature",
                args,
                env_vars,
            )

    def _run_remove_signatures(self, signatures_to_remove: List[str]) -> None:
        """Remove signatures from the sigstore.

        Args:
            signatures_to_remove (List[str]): List of signatures to remove.
        """
        cert, key = self.settings.pyxis_ssl_crt_file, self.settings.pyxis_ssl_key_file
        args = []
        args = ["--pyxis-server", self.settings.pyxis_server]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]

        with tempfile.NamedTemporaryFile(mode="w") as temp:
            json.dump(signatures_to_remove, temp)
            temp.flush()

            args += ["--ids", "@%s" % temp.name]

    def _filter_to_remove(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> List[str]:
        """Filter signatures to remove.

        Args:
            signatures (List[Tuple[str, str, str]]): List of (digest, tag, repository)
            tuples of signautres to remove.
            _exclude (Optional[List[Tuple[str, str, str]]]): List of  (digest, tag, repository)
            tuples of signautres to keep.
        """
        exclude = _exclude or []
        signatures_to_remove = list(self._fetch_signatures([x[0] for x in signatures]))
        sig_ids_to_remove = []
        for existing_signature in signatures_to_remove:
            if (
                existing_signature["manifest_digest"],
                existing_signature["reference"].split(":")[-1],
                existing_signature["repository"],
            ) in signatures and (
                existing_signature["manifest_digest"],
                existing_signature["reference"],
                existing_signature["repository"],
            ) not in exclude:
                sig_ids_to_remove.append(existing_signature["_id"])
                LOG.debug(
                    f"Removing signature. Reference: {existing_signature['reference']}, "
                    f"Repository: {existing_signature['repository']}, "
                    f"Digest: {existing_signature['manifest_digest']}, "
                    f"Key: {existing_signature['sig_key_id']}"
                )
        return sig_ids_to_remove

    def remove_signatures(
        self,
        signatures: List[Tuple[str, str, str]],
        _exclude: Optional[List[Tuple[str, str, str]]] = None,
    ) -> None:
        """Remove signatures from sigstore.

        Args:
            signatures (list): List of tuples containing (digest, reference, repository) of
            signatures to remove.
            exclude (Optional[List[Tuple[str, str, str]]]): List of  (digest, tag, repository)
            tuples of signautres to keep.
        """
        _signatures = list(signatures)
        to_remove = self._filter_to_remove(_signatures, _exclude=_exclude)
        self._remove_signatures(to_remove)


class CosignSignerSettings(Base):
    """Validation schema for cosign signer settings."""


class CosignSignerWrapper(SignerWrapper):
    """Wrapper for cosign signer functionality."""

    label: str = dataclasses.field(default="cosign_signer", init=False)
    pre_push: bool = dataclasses.field(default=False, init=False)
    settings: CosignSignerSettings = dataclasses.field(default_factory=CosignSignerSettings)

    _entry_point_conf = ["pubtools-sign", "modules", "pubtools-sign-cosign-container-sign"]


SIGNER_BY_LABEL = {
    wrapper.label: wrapper
    for name, wrapper in locals().items()
    if type(wrapper) is type and issubclass(wrapper, SignerWrapper) and wrapper != SignerWrapper
}



from dataclasses import field

from typing import Any, Union

from pytractions.base import TDict, TList, Base, doc
from signtractions.resources.signing_wrapper import (
    CosignSignerWrapper,
)


class FakeEPRunArgs(Base):
    """Fake entry point run args for testing."""

    args: TList[str]
    kwargs: TDict[str, Union[str, TList[str]]]


class FakeCosignSignerWrapper(CosignSignerWrapper):
    """Fake cosign signer wrapper for testing."""

    fake_entry_point_requests: TList[FakeEPRunArgs]
    fake_entry_point_returns: TList[TDict[str, TDict[str, str]]]
    fake_entry_point_runs: TList[FakeEPRunArgs] = field(default_factory=TList[FakeEPRunArgs])

    d_fake_entry_point_requests: str = doc("Fake entry point requests for testing.")
    d_fake_entry_point_returns: str = doc("Fake entry point returns for testing.")
    d_fake_entry_point_runs: str = doc("Records of fake entry point runs.")

    def _fake_ep(self, *args, **kwargs):
        self.fake_entry_point_runs.append(
            FakeEPRunArgs(
                args=TList[str](args),
                kwargs=TDict[str, Union[str, TList[str]]].content_from_json(kwargs),
            )
        )

        i = self.fake_entry_point_requests.index(
            FakeEPRunArgs(
                args=TList[str](args),
                kwargs=TDict[str, Union[str, TList[str]]].content_from_json(kwargs),
            )
        )
        return self.fake_entry_point_returns[i]

    @property
    def entry_point(self) -> Any:
        """Load and return entry point for pubtools-sign project."""
        return self._fake_ep



import pytest
from typing import Union

from pytractions.base import In, TList, TDict, Res

from signtractions.resources.signing_wrapper import CosignSignerSettings
from signtractions.resources.fake_signing_wrapper import FakeCosignSignerWrapper, FakeEPRunArgs
from signtractions.tractions.signing import (
    SignEntriesFromContainerParts,
    SignEntry,
    ContainerParts,
    SignSignEntries,
)
from signtractions.resources.signing_wrapper import SigningError


def test_sign_entries_from_container_parts():
    t = SignEntriesFromContainerParts(
        uid="test",
        i_container_parts=In[ContainerParts](
            data=ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="latest",
                digests=TList[str](["sha256:123456", "sha256:123457"]),
                arches=TList[str](["amd64", "arm64"]),
            )
        ),
        i_signing_key=In[str](data="signing_key"),
    )
    t.run()
    assert t.o_sign_entries.data[0] == SignEntry(
        repo="containers/podman",
        reference="quay.io/containers/podman:latest",
        digest="sha256:123456",
        arch="amd64",
        signing_key="signing_key",
    )
    assert t.o_sign_entries.data[1] == SignEntry(
        repo="containers/podman",
        reference="quay.io/containers/podman:latest",
        digest="sha256:123457",
        arch="arm64",
        signing_key="signing_key",
    )


def test_sign_sign_entries():
    fsw = FakeCosignSignerWrapper(
        config_file="test",
        settings=CosignSignerSettings(),
        fake_entry_point_requests=TList[FakeEPRunArgs]([]),
        fake_entry_point_returns=TList[TDict[str, TDict[str, str]]]([]),
        fake_entry_point_runs=TList[FakeEPRunArgs]([]),
    )
    fsw.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]](
                {
                    "config_file": "test",
                    "signing_key": "signing_key",
                    "digest": TList[str](["sha256:123456"]),
                    "reference": TList[str](["quay.io/containers/podman:latest"]),
                }
            ),
        )
    )
    fsw.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]]({"signer_result": TDict[str, str]({"status": "ok"})})
    )
    t = SignSignEntries(
        uid="test",
        r_signer_wrapper=Res[FakeCosignSignerWrapper](r=fsw),
        i_task_id=In[int](data=1),
        i_sign_entries=In[TList[SignEntry]](
            data=TList[SignEntry](
                [
                    SignEntry(
                        repo="containers/podman",
                        reference="quay.io/containers/podman:latest",
                        digest="sha256:123456",
                        arch="amd64",
                        signing_key="signing_key",
                    )
                ]
            )
        ),
    )
    t.run()


def test_sign_sign_entries_fail():
    fsw = FakeCosignSignerWrapper(
        config_file="test",
        settings=CosignSignerSettings(),
        fake_entry_point_requests=TList[FakeEPRunArgs]([]),
        fake_entry_point_returns=TList[TDict[str, TDict[str, str]]]([]),
        fake_entry_point_runs=TList[FakeEPRunArgs]([]),
    )
    fsw.fake_entry_point_requests.append(
        FakeEPRunArgs(
            args=TList[str]([]),
            kwargs=TDict[str, Union[str, TList[str]]].content_from_json(
                {
                    "config_file": "test",
                    "digest": TList[str](["sha256:123456"]),
                    "reference": TList[str](["quay.io/containers/podman:latest"]),
                    "signing_key": "signing_key",
                }
            ),
        )
    )
    fsw.fake_entry_point_returns.append(
        TDict[str, TDict[str, str]].content_from_json(
            {"signer_result": {"status": "error", "error_message": "test error"}}
        )
    )
    t = SignSignEntries(
        uid="test",
        r_signer_wrapper=Res[FakeCosignSignerWrapper](r=fsw),
        i_task_id=In[int](data=1),
        i_sign_entries=In[TList[SignEntry]](
            data=TList[SignEntry](
                [
                    SignEntry(
                        repo="containers/podman",
                        reference="quay.io/containers/podman:latest",
                        digest="sha256:123456",
                        arch="amd64",
                        signing_key="signing_key",
                    )
                ]
            )
        ),
    )
    with pytest.raises(SigningError):
        t.run()



import json
import hashlib
import logging
from typing import Type
from pytractions.base import Traction, TList, In, Out, Res, STMD, OnUpdateCallable

from ..models.containers import ContainerParts

from ..resources.quay_client import QuayClient

LOG = logging.getLogger()
logging.basicConfig()
LOG.setLevel(logging.INFO)


class ParseCotainerImageReference(Traction):
    """Parser container image reference into parts."""

    i_container_image_reference: In[str]
    o_container_parts: Out[ContainerParts]

    d_: str = """Parser container image reference into ContainerParts model"""
    d_i_container_image_reference: str = "Container image reference to parse"
    d_o_container_parts: str = "Parsed container parts"

    def _run(self, on_update: OnUpdateCallable = None) -> None:
        registry, rest = self.i_container_image_reference.data.split("/", 1)
        if "@" in rest:
            image, digest = rest.split("@", 1)
            tag = None
        else:
            image, tag = rest.split(":", 1)
            digest = None
        self.o_container_parts.data = ContainerParts(
            registry=registry,
            image=image,
            tag=tag,
            digests=TList[str]([digest]) if digest else TList[str](),
            arches=TList[str]([""]) if digest else TList[str](),
        )
        self.add_details("parsed container parts" + str(self.o_container_parts.data))


class STMDParseContainerImageReference(STMD):
    """Parser container image references into list of parts."""

    _traction: Type[Traction] = ParseCotainerImageReference
    i_container_image_reference: In[TList[str]]
    o_container_parts: Out[TList[ContainerParts]]

    d_: str = """Parser container image reference into ContainerParts model. STMD version."""
    d_i_container_image_reference: str = "List of container image references to parse"
    d_o_container_parts: str = "List of Parsed container parts"


class PopulateContainerDigest(Traction):
    """Fetch digest(s) for ContainerParts if there isn't any."""

    i_container_parts: In[ContainerParts]
    o_container_parts: Out[ContainerParts]
    r_quay_client: Res[QuayClient]

    d_: str = """Fetch digest(s) for ContainerParts if there aren't any

    If fetched manifest by tag is manifest lists, populate also digests for manifests in the
    manifest list + digest of the list itself.
    """
    d_i_container_parts: str = "Container parts to fetch digest for"
    d_o_container_parts: str = "Container parts with digests populated (or unchanged)"
    d_r_quay_client: str = "Quay client to fetch manifest"

    def _run(self, on_update: OnUpdateCallable = None) -> None:
        self.o_container_parts.data = self.i_container_parts.data
        if self.i_container_parts.data.tag:
            LOG.info(
                "Fetching {}/{}:{}".format(
                    self.i_container_parts.data.registry,
                    self.i_container_parts.data.image,
                    self.i_container_parts.data.tag,
                )
            )
            manifest_str = self.r_quay_client.r.get_manifest(
                "{}/{}:{}".format(
                    self.i_container_parts.data.registry,
                    self.i_container_parts.data.image,
                    self.i_container_parts.data.tag,
                ),
                raw=True,
            )
        else:
            LOG.info(
                "Fetching {}/{}@{}".format(
                    self.i_container_parts.data.registry,
                    self.i_container_parts.data.image,
                    self.i_container_parts.data.digests[0],
                )
            )
            manifest_str = self.r_quay_client.r.get_manifest(
                "{}/{}@{}".format(
                    self.i_container_parts.data.registry,
                    self.i_container_parts.data.image,
                    self.i_container_parts.data.digests[0],
                ),
                raw=True,
            )

        manifest = json.loads(manifest_str)
        self.o_container_parts.data = ContainerParts(
            registry=self.i_container_parts.data.registry,
            image=self.i_container_parts.data.image,
            tag=self.i_container_parts.data.tag,
        )
        if manifest["mediaType"] in (
            QuayClient._MANIFEST_LIST_TYPE,
            QuayClient._MANIFEST_OCI_LIST_TYPE,
        ):
            for _manifest in manifest["manifests"]:
                self.o_container_parts.data.digests.append(_manifest["digest"])
                self.o_container_parts.data.arches.append(_manifest["platform"]["architecture"])

            hasher = hashlib.sha256()
            hasher.update(manifest_str.encode("utf-8"))
            digest = hasher.hexdigest()
            self.o_container_parts.data.digests.append("sha256:" + digest)
            self.o_container_parts.data.arches.append("multiarch")

        else:
            hasher = hashlib.sha256()
            hasher.update(manifest_str.encode("utf-8"))
            digest = hasher.hexdigest()
            self.o_container_parts.data.digests.append("sha256:" + digest)
            self.o_container_parts.data.arches.append("")


class STMDPopulateContainerDigest(STMD):
    """Fetch digest(s) for ContainerParts if there isn't any. STMD version."""

    _traction: Type[Traction] = PopulateContainerDigest
    i_container_parts: In[TList[ContainerParts]]
    o_container_parts: Out[TList[ContainerParts]]
    r_quay_client: Res[QuayClient]

    d_: str = """Fetch digest(s) for ContainerParts if there aren't any

    If fetched manifest by tag is manifest lists, populate also digests for manifests in the
    manifest list + digest of the list itself.
    """
    d_i_container_parts: str = "Container parts to fetch digest for"
    d_o_container_parts: str = "Container parts with digests populated (or unchanged)"
    d_r_quay_client: str = "Quay client to fetch manifest"



from typing import Union

from pytractions.base import TList, Arg, Res, TRes, In, TIn, Out, STMDExecutorType
from pytractions.transformations import Flatten
from pytractions.tractor import Tractor

from ..tractions.containers import STMDPopulateContainerDigest, STMDParseContainerImageReference
from ..tractions.signing import SignSignEntries
from ..tractions.signing import STMDSignEntriesFromContainerParts

from ..resources.quay_client import QuayClient
from ..resources.signing_wrapper import MsgSignerWrapper, CosignSignerWrapper
from ..resources.fake_signing_wrapper import FakeCosignSignerWrapper
from ..resources.fake_quay_client import FakeQuayClient

from ..models.signing import SignEntry


class SignContainers(Tractor):
    """Sign container images."""

    r_signer_wrapper_cosign: Res[
        Union[MsgSignerWrapper, CosignSignerWrapper, FakeCosignSignerWrapper]
    ] = TRes[Union[MsgSignerWrapper, CosignSignerWrapper, FakeCosignSignerWrapper]]()
    r_dst_quay_client: Res[Union[QuayClient, FakeQuayClient]] = TRes[
        Union[QuayClient, FakeQuayClient]
    ]()
    i_container_image_references: In[TList[str]] = TIn[TList[str]]()
    i_task_id: In[int] = In[int](data=1)
    i_signing_keys: In[TList[str]] = TIn[TList[str]]()
    a_pool_size: Arg[int] = Arg[int](a=10)
    a_executor_type: Arg[STMDExecutorType] = Arg[STMDExecutorType](a=STMDExecutorType.THREAD)

    t_parse_container_references: STMDParseContainerImageReference = (
        STMDParseContainerImageReference(
            uid="parse_container_references",
            i_container_image_reference=i_container_image_references,
            a_pool_size=a_pool_size,
            a_executor_type=a_executor_type,
        )
    )
    t_populate_digests: STMDPopulateContainerDigest = STMDPopulateContainerDigest(
        uid="populate_digests",
        r_quay_client=r_dst_quay_client,
        i_container_parts=t_parse_container_references.o_container_parts,
        a_pool_size=a_pool_size,
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.THREAD),
    )
    t_sign_entries_from_push_item: STMDSignEntriesFromContainerParts = (
        STMDSignEntriesFromContainerParts(
            uid="sign_entries_from_container_parts",
            i_container_parts=t_populate_digests.o_container_parts,
            i_signing_key=i_signing_keys,
            a_pool_size=a_pool_size,
            a_executor_type=a_executor_type,
        )
    )
    t_flatten_sign_entries: Flatten[SignEntry] = Flatten[SignEntry](
        uid="flatten_sign_entries",
        i_complex=t_sign_entries_from_push_item.o_sign_entries,
    )
    t_sign_push_item_after_push: SignSignEntries = SignSignEntries(
        uid="sign_push_item_after_push",
        i_sign_entries=t_flatten_sign_entries.o_flat,
        i_task_id=i_task_id,
        r_signer_wrapper=r_signer_wrapper_cosign,
    )
    o_sign_entries: Out[TList[SignEntry]] = t_flatten_sign_entries.o_flat

    d_: str = """
    Sign container images provided as input in cosign wrapper with signing keys provided as input.
"""
    d_i_task_id: str = "Task ID to identify signing request."
    d_i_container_image_references: str = "List of container image references to sign."
    d_a_pool_size: str = "Pool size used for STMD tractions"
    d_i_signing_keys: str = "List of signing keys used to sign containers. One key per container."



from signtractions.resources.reference_constructor import (
    ReferenceConstructorInternal,
    ReferenceConstructorExternal,
)


def test_reference_constructor_external():
    """Test ReferenceConstructorExternal."""
    rce = ReferenceConstructorExternal(external_type=1)
    assert rce.construct_reference_tag("registry", "test/repo", "tag") == "registry/test/repo:tag"
    assert (
        rce.construct_reference_digest("registry", "test/repo", "digest")
        == "registry/test/repo@digest"
    )


def test_reference_constructor_internal():
    """Test ReferenceConstructorInternal."""
    rce = ReferenceConstructorInternal(quay_org="quay_org")
    assert (
        rce.construct_reference_tag("registry", "test/repo", "tag")
        == "registry/quay_org/test----repo:tag"
    )
    assert (
        rce.construct_reference_digest("registry", "test/repo", "digest")
        == "registry/quay_org/test----repo@digest"
    )



import json
import pytest

from pytractions.base import Arg, In, TList, TDict, STMDExecutorType, Res

from signtractions.resources.fake_quay_client import FakeQuayClient, ManifestNotFoundError

from signtractions.tractions.containers import (
    ParseCotainerImageReference,
    PopulateContainerDigest,
    STMDParseContainerImageReference,
    ContainerParts,
)


def test_parse_container_image_reference_tag():
    t = ParseCotainerImageReference(
        uid="test",
        i_container_image_reference=In[str](data="quay.io/containers/podman:latest"),
    )
    t.run()
    assert t.o_container_parts.data.registry == "quay.io"
    assert t.o_container_parts.data.image == "containers/podman"
    assert t.o_container_parts.data.tag == "latest"
    assert t.o_container_parts.data.digests == TList[str]([])
    assert t.o_container_parts.data.arches == TList[str]([])


def test_parse_container_image_reference_digest():
    t = ParseCotainerImageReference(
        uid="test",
        i_container_image_reference=In[str](data="quay.io/containers/podman@sha256:123456"),
    )
    t.run()
    assert t.o_container_parts.data.registry == "quay.io"
    assert t.o_container_parts.data.image == "containers/podman"
    assert t.o_container_parts.data.tag is None
    assert t.o_container_parts.data.digests == TList[str](["sha256:123456"])
    assert t.o_container_parts.data.arches == TList[str]([""])


def test_parse_container_image_reference_stmd_tag():
    t = STMDParseContainerImageReference(
        uid="test",
        i_container_image_reference=In[TList[str]](
            data=TList[str](
                [
                    "quay.io/containers/podman:latest",
                    "quay.io/containers/podman:greatest",
                ]
            )
        ),
        a_executor_type=Arg[STMDExecutorType](a=STMDExecutorType.LOCAL),
        a_pool_size=Arg[int](a=1),
    )
    t.run()
    assert t.o_container_parts.data == TList[ContainerParts](
        [
            ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="latest",
                digests=TList[str]([]),
                arches=TList[str]([]),
            ),
            ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="greatest",
                digests=TList[str]([]),
                arches=TList[str]([]),
            ),
        ]
    )


def test_populate_container_digest_manifest_not_found():
    fqc = FakeQuayClient(
        username="test",
        password="test",
        host="test",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    t = PopulateContainerDigest(
        uid="test",
        i_container_parts=In[ContainerParts](
            data=ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="latest",
                digests=TList[str]([]),
                arches=TList[str]([]),
            )
        ),
        r_quay_client=Res[FakeQuayClient](r=fqc),
    )
    with pytest.raises(ManifestNotFoundError):
        t.run()


def test_populate_container_digest_manifest(fix_manifest_v2s2):
    fqc = FakeQuayClient(
        username="test",
        password="test",
        host="test",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    fqc.populate_manifest(
        "quay.io/containers/podman:latest",
        "application/vnd.docker.distribution.manifest.v2+json",
        {},
        json.dumps(fix_manifest_v2s2),
    )

    t = PopulateContainerDigest(
        uid="test",
        i_container_parts=In[ContainerParts](
            data=ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="latest",
                digests=TList[str]([]),
                arches=TList[str]([]),
            )
        ),
        r_quay_client=Res[FakeQuayClient](r=fqc),
    )
    t.run()
    assert t.o_container_parts.data.registry == "quay.io"
    assert t.o_container_parts.data.image == "containers/podman"
    assert t.o_container_parts.data.tag == "latest"
    assert t.o_container_parts.data.digests == TList[str](
        ["sha256:6ef06d8c90c863ba4eb4297f1073ba8cb28c1f6570e2206cdaad2084e2a4715d"]
    )


def test_populate_container_digest_manifest_list(fix_manifest_list):
    fqc = FakeQuayClient(
        username="test",
        password="test",
        host="test",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    fqc.populate_manifest(
        "quay.io/containers/podman:latest",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        {},
        json.dumps(fix_manifest_list),
    )

    t = PopulateContainerDigest(
        uid="test",
        i_container_parts=In[ContainerParts](
            data=ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="latest",
                digests=TList[str]([]),
                arches=TList[str]([]),
            )
        ),
        r_quay_client=Res[FakeQuayClient](r=fqc),
    )
    t.run()
    assert t.o_container_parts.data.registry == "quay.io"
    assert t.o_container_parts.data.image == "containers/podman"
    assert t.o_container_parts.data.tag == "latest"
    assert t.o_container_parts.data.digests == TList[str](
        [
            "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
            "sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3",
        ]
    )


def test_populate_container_digest_manifest_list_by_digest(fix_manifest_list):
    fqc = FakeQuayClient(
        username="test",
        password="test",
        host="test",
        fake_manifests=TDict[str, TDict[str, str]].content_from_json({}),
    )
    fqc.populate_manifest(
        "quay.io/containers/podman@sha256:2e8f38a0a8d2a450598430fa"
        "70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
        "application/vnd.docker.distribution.manifest.list.v2+json",
        {},
        json.dumps(fix_manifest_list),
    )

    t = PopulateContainerDigest(
        uid="test",
        i_container_parts=In[ContainerParts](
            data=ContainerParts(
                registry="quay.io",
                image="containers/podman",
                tag="",
                digests=TList[str](
                    ["sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee"]
                ),
                arches=TList[str]([]),
            )
        ),
        r_quay_client=Res[FakeQuayClient](r=fqc),
    )
    t.run()
    assert t.o_container_parts.data.registry == "quay.io"
    assert t.o_container_parts.data.image == "containers/podman"
    assert t.o_container_parts.data.tag == ""
    assert t.o_container_parts.data.digests == TList[str](
        [
            "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
            "sha256:d07476154b88059d730e260eba282b3c7a0b5e7feb620638d49070b71dcdcaf3",
        ]
    )



import json
import logging
import shlex
import subprocess
import textwrap
from types import TracebackType
from typing import Any, Optional, Type
from typing_extensions import Self


from shlex import quote

# from pubtools.tracing import get_trace_wrapper

# tw = get_trace_wrapper()
LOG = logging.getLogger("pubtools.quay")


class Executor(object):
    """
    Base executor class.

    Implementation of command execution should be done in
    descendant classes. Common pre- and post-processing should be
    implemented in this class.
    """

    def __enter__(self) -> Self:
        """Use the class as context manager. Returns instance upon invocation."""
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Cleanup when used as context manager. No-op by default."""
        pass

    def _run_cmd(
        self,
        cmd: str,
        err_msg: str | None = None,
        tolerate_err: bool = False,
        stdin: str | None = None,
    ) -> tuple[str, str]:
        """Run a bash command."""
        raise NotImplementedError  # pragma: no cover"

    def skopeo_login(
        self, host: str = "quay.io", username: str | None = None, password: str | None = None
    ) -> None:
        """
        Attempt to login to Quay if no login credentials are present.

        Args:
            host (str):
                docker registry host (quay.io as default)
            username (str):
                Username for login.
            password (str):
                Password for login.
        """
        cmd_check = "skopeo login --get-login %s" % host
        out, err = self._run_cmd(cmd_check, tolerate_err=True)
        if username and username in out:
            LOG.info("Already logged in to Quay.io")
            return

        if not username or not password:
            raise ValueError(
                "Skopeo login credentials are not present. Quay user and token must be provided."
            )
        LOG.info(f"Logging in to Quay with provided credentials {username} {password}")

        cmd_login = ("skopeo login -u {0} --password-stdin %s" % host).format(quote(username))
        out, err = self._run_cmd(cmd_login, stdin=password)

        if "Login Succeeded" in out:
            LOG.info("Login successful")
        else:
            raise RuntimeError(
                "Login command didn't generate expected output. "
                "STDOUT: '{0}', STDERR: '{1}'".format(out, err)
            )

    def tag_images(self, source_ref: str, dest_refs: list[str], all_arch: bool = False) -> None:
        """
        Copy image from source to destination(s) using skopeo.

        Args:
            source_ref (str):
                Reference of the source image.
            dest_refs ([str]):
                List of target references to copy the image to.
            all_arch (bool):
                Whether to copy all architectures (if multiarch image)
        """
        if all_arch:
            cmd = "skopeo copy --all docker://{0} docker://{1} --format v2s2"
        else:
            cmd = "skopeo copy docker://{0} docker://{1} --format v2s2"

        for dest_ref in dest_refs:
            LOG.info("Tagging source '{0}' to destination '{1}'".format(source_ref, dest_ref))
            self._run_cmd(cmd.format(quote(source_ref), quote(dest_ref)))
            LOG.info("Destination image {0} has been tagged.".format(dest_ref))

        LOG.info("Tagging complete.")

    def skopeo_inspect(self, image_ref: str, raw: bool = False) -> Any:
        """
        Run skopeo inspect and return the result.

        NOTE: inspect command will not be run with the --raw argument. This option only returns an
        image manifest, which can be gathered by QuayClient. 'raw' argument in this function
        denotes if the result should be parsed or returned raw.

        Args:
            image_ref (str):
                Image reference to inspect.
            raw (bool):
                Whether to parse the returned JSON, or return raw.
        Returns (dict|str):
            Result of skopeo inspect.
        """
        cmd = "skopeo inspect docker://{0}".format(image_ref)
        out, _ = self._run_cmd(cmd)

        if raw:
            return out
        else:
            return json.loads(out)


class LocalExecutor(Executor):
    """Run commands locally."""

    def __init__(self, params: dict[str, Any] = {}) -> None:
        """
        Initialize.

        Args:
            params (dict):
                Custom parameters to be applied when running the shell commands.
        """
        self.params = params
        self.params.setdefault("universal_newlines", True)
        self.params.setdefault("stderr", subprocess.PIPE)
        self.params.setdefault("stdout", subprocess.PIPE)
        self.params.setdefault("stdin", subprocess.PIPE)

    # @tw.instrument_func(args_to_attr=True)
    def _run_cmd(
        self,
        cmd: str,
        err_msg: str | None = None,
        tolerate_err: bool = False,
        stdin: str | None = None,
    ) -> tuple[str, str]:
        """
        Run a command locally.

        Args:
            cmd (str):
                Shell command to be executed.
            error_msg (str):
                Error message written when the command fails.
            tolerate_err (bool):
                Whether to tolerate a failed command.
            stdin (str):
                String to send to standard input for a command.

        Returns (str, str):
            Tuple of stdout and stderr generated by the command.
        """
        err_msg = err_msg or "An error has occured when executing a command."

        p = subprocess.Popen(shlex.split(cmd), **self.params)
        out, err = p.communicate(input=stdin)

        if p.returncode != 0 and not tolerate_err:
            LOG.error("Command {0} failed with the following error:".format(cmd))
            for line in textwrap.wrap(err, 200):
                LOG.error(f"    {line}")
            raise RuntimeError(err_msg)

        return out, err



from pytractions.base import Base


class ReferenceConstructor(Base):
    """Container reference constructor."""

    def construct_reference_digest(self, registry, repo, digest):
        """Construct docker reference for registry repo and digest."""
        pass  # pragma: no cover

    def construct_reference_tag(self, registry, repo, tag):
        """Construct docker reference for registry repo and tag."""


class ReferenceConstructorInternal(ReferenceConstructor):
    """Container reference constructor for repos in internal format."""

    quay_org: str

    def construct_reference_tag(self, registry, repo, tag):
        """Construct docker reference for registry repo and tag."""
        return f"{registry}/{self.quay_org}/{repo.replace('/','----')}:{tag}"

    def construct_reference_digest(self, registry, repo, digest):
        """Construct docker reference for registry repo and digest."""
        return f"{registry}/{self.quay_org}/{repo.replace('/','----')}@{digest}"


class ReferenceConstructorExternal(ReferenceConstructor):
    """Container reference constructor for repos in external format."""

    external_type: int

    def construct_reference_tag(self, registry, repo, tag):
        """Construct docker reference for registry repo and tag."""
        return f"{registry}/{repo}:{tag}"

    def construct_reference_digest(self, registry, repo, digest):
        """Construct docker reference for registry repo and digest."""
        return f"{registry}/{repo}@{digest}"



from typing import Optional

from dataclasses import field

from pytractions.base import Base, TList, TDict


class ComponentSourceGit(Base):
    """Component source data structure."""

    context: str = ""
    dockerfileUrl: str = ""
    revision: str = ""
    url: str = ""


class ComponentSource(Base):
    """Component source data structure."""

    git: Optional[ComponentSourceGit]


class SnapshotSpecComponent(Base):
    """Snapshot component data structure."""

    name: str = ""
    containerImage: str = ""
    repository: str = ""
    source: Optional[ComponentSource] = None


class SnapshotSpec(Base):
    """Data structure to hold container reference parts."""

    application: str = ""
    components: TList[SnapshotSpecComponent] = field(default_factory=TList[SnapshotSpecComponent])
    artifacts: TDict[str, str] = field(default_factory=TDict[str, str])



from typing import Type

from pytractions.base import (
    Traction,
    STMD,
    In,
    Out,
    Res,
    OnUpdateCallable,
    TList,
    STMDSingleIn,
)
from ..resources.signing_wrapper import SignerWrapper
from ..models.signing import SignEntry
from ..models.containers import ContainerParts


class SignSignEntries(Traction):
    """Sign SignEntries."""

    r_signer_wrapper: Res[SignerWrapper]
    i_task_id: In[int]
    i_sign_entries: In[TList[SignEntry]]

    d_: str = "Sign provided SignEntries with signer wrapper."
    d_i_sign_entries: str = "List of SignEntry objects to sign."
    d_i_task_id: str = "Task id used to identify signing requests."

    def _run(self, on_update: OnUpdateCallable = None) -> None:
        self.r_signer_wrapper.r.sign_containers(
            [x for x in self.i_sign_entries.data],
            task_id=self.i_task_id.data,
        )


class STMDSignSignEntries(STMD):
    """Sign SignEntries.

    STMD version
    """

    _traction: Type[Traction] = SignSignEntries
    r_signer_wrapper: Res[SignerWrapper]
    i_task_id: STMDSingleIn[int]
    i_sign_entries: In[TList[TList[SignEntry]]]

    d_: str = "Sign provided SignEntries with signer wrapper. STMD version."
    d_i_sign_entries: str = "List of List of SignEntry objects to sign."
    d_i_task_id: str = "Task id used to identify signing requests."


class SignEntriesFromContainerParts(Traction):
    """Create sign entries from container parts."""

    i_container_parts: In[ContainerParts]
    i_signing_key: In[str]
    o_sign_entries: Out[TList[SignEntry]]

    d_: str = """Create sign entries from container parts.
    For each pair of digest and arch in container parts, create a SignEntry object.
    """
    d_i_signing_key: str = "Signing key to use for signing."
    d_i_container_parts: str = "Container parts to create sign entries from."
    d_o_sign_entries: str = "List of SignEntry objects"

    def _run(self, on_update: OnUpdateCallable = None) -> None:
        for digest, arch in zip(
            self.i_container_parts.data.digests, self.i_container_parts.data.arches
        ):
            self.o_sign_entries.data.append(
                SignEntry(
                    digest=digest,
                    arch=arch,
                    reference=(
                        self.i_container_parts.data.make_reference()
                        if self.i_container_parts.data.tag
                        else None
                    ),
                    repo=self.i_container_parts.data.image,
                    signing_key=self.i_signing_key.data,
                )
            )


class STMDSignEntriesFromContainerParts(STMD):
    """Create sign entries from container parts.

    STMD version
    """

    _traction: Type[Traction] = SignEntriesFromContainerParts
    i_signing_key: In[TList[str]]
    i_container_parts: In[TList[ContainerParts]]
    o_sign_entries: Out[TList[TList[SignEntry]]]

    d_: str = """Create sign entries from container parts.
    For each pair of digest and arch in container parts, create a SignEntry object.
    STMD version
    """
    d_i_signing_key: str = "List of signing key to use for signing."
    d_i_container_parts: str = "List of container parts to create sign entries from."
    d_o_sign_entries: str = "List of List of SignEntry objects"



# -*- coding: utf-8 -*-

"""setup.py"""

from setuptools import setup, find_packages

from pytractions.pkgutils import traction_entry_points

import signtractions.tractors.t_sign_containers
import signtractions.tractors.t_sign_snapshot

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]


def get_requirements():
    """
    Transform a list of requirements so that they are usable by older pip (9.0.0), and newer pip

    Regex extracts name and url from a tox-compatible format, and replaces it with only a name
    (which will be combined with dependency_links) or with PEP-508 compatible dependency.
    """
    with open("requirements.txt") as f:
        reqs = f.read().splitlines()
        return reqs


setup(
    name="signtractions",
    version="0.0.4",
    description="quay-tractions",
    long_description="pubtools-sign pytractions wrapper for signing containers and release snapshots in tractions.",
    long_description_content_type="text/x-rst",
    author="Jindrich Luza",
    author_email="jluza@redhat.com",
    url="https://github.com/midnightercz/signtractions",
    classifiers=classifiers,
    python_requires=">=3.6",
    packages=find_packages(exclude=["tests"]),
    data_files=[],
    install_requires=[
        "koji",
        "requests",
        "requests_kerberos",
        "pytractions",
        "gssapi"],
    entry_points={
        "tractions": list(set([
            x for x in traction_entry_points(signtractions.tractors.t_sign_containers)
        ]) & set([
            x for x in traction_entry_points(signtractions.tractors.t_sign_snapshot)
        ])),
    },
    dependency_links=[
    ],
    include_package_data=True,
)



from typing import Union

from pytractions.base import TList, Arg, Res, TRes, In, TIn, Out, STMDExecutorType
from pytractions.transformations import ListMultiplier
from pytractions.tractor import Tractor

from ..tractions.snapshot import ParseSnapshot, ContainerImagesFromSnapshot

from ..resources.quay_client import QuayClient
from ..resources.fake_quay_client import FakeQuayClient

from ..resources.signing_wrapper import MsgSignerWrapper, CosignSignerWrapper
from ..resources.fake_signing_wrapper import FakeCosignSignerWrapper

from ..models.signing import SignEntry

from .t_sign_containers import SignContainers


class SignSnapshot(Tractor):
    """Sign release snapshot."""

    r_signer_wrapper_cosign: Res[
        Union[FakeCosignSignerWrapper, MsgSignerWrapper, CosignSignerWrapper]
    ] = TRes[Union[FakeCosignSignerWrapper, MsgSignerWrapper, CosignSignerWrapper]]()
    r_dst_quay_client: Res[Union[QuayClient, FakeQuayClient]] = TRes[
        Union[QuayClient, FakeQuayClient]
    ]()
    i_snapshot_str: In[str] = TIn[str]()
    i_snapshot_file: In[str] = TIn[str]()
    i_signing_key: In[str] = TIn[str]()
    i_task_id: In[int] = TIn[int]()
    a_pool_size: Arg[int] = Arg[int](a=10)
    a_executor_type: Arg[STMDExecutorType] = Arg[STMDExecutorType](a=STMDExecutorType.THREAD)

    t_parse_snapshot: ParseSnapshot = ParseSnapshot(
        uid="parse_snapshot",
        i_snapshot_str=i_snapshot_str,
        i_snapshot_file=i_snapshot_file,
    )
    t_container_images_from_snapshot: ContainerImagesFromSnapshot = ContainerImagesFromSnapshot(
        uid="container_images_from_snapshot", i_snapshot_spec=t_parse_snapshot.o_snapshot_spec
    )
    t_populate_signing_keys: ListMultiplier[str, str] = ListMultiplier[str, str](
        uid="populate_signing_keys",
        i_scalar=i_signing_key,
        i_list=t_container_images_from_snapshot.o_container_images,
    )
    t_sign_containers: SignContainers = SignContainers(
        uid="sign_containers",
        r_signer_wrapper_cosign=r_signer_wrapper_cosign,
        r_dst_quay_client=r_dst_quay_client,
        i_container_image_references=t_container_images_from_snapshot.o_container_images,
        i_signing_keys=t_populate_signing_keys.o_list,
        i_task_id=i_task_id,
        a_pool_size=Arg[int](a=10),
        a_executor_type=a_executor_type,
    )

    o_sign_entries: Out[TList[SignEntry]] = t_sign_containers.o_sign_entries

    d_: str = """
    Sign containers in release snapshot.
"""
    d_i_task_id: str = "Task ID to identify signing request."
    d_i_snapshot_str: str = "Json representation of release snapshot."
    d_i_snapshot_file: str = "Path to a file containing snapshot in json format."
    d_a_pool_size: str = "Pool size used for STMD tractions."
    d_a_executor_type: str = "STMD executor type."
    d_i_signing_key: str = "Signing key used to sign containers. One key per container."
    d_r_signer_wrapper_cosign: str = "Signer wrapper used to sign container images."
    d_r_dst_quay_client: str = (
        "Quay client used for fetching container images when populating " "digests in SignEntries."
    )
    d_t_container_images_from_snapshot: str = "Extract container images from snapshot."
    d_t_populate_signing_keys: str = "Populate signing keys for each container image."
    d_t_sign_containers: str = "Sign containers in release snapshot."



import json

from pytractions.base import (
    Traction,
    In,
    Out,
    TList,
    OnUpdateCallable,
)
from ..models.snapshot import SnapshotSpec


class ParseSnapshot(Traction):
    """Sign SignEntries."""

    i_snapshot_str: In[str]
    i_snapshot_file: In[str]
    o_snapshot_spec: Out[SnapshotSpec]

    d_: str = "Parse snapshot json string into Snapshot object."
    d_i_snapshot_str: str = "Snapshot string"
    d_o_snapshot_spec: str = "Parsed Snapshot object"

    def _run(self, on_update: OnUpdateCallable = None) -> None:
        if self.i_snapshot_str.data:
            self.o_snapshot_spec.data = SnapshotSpec.content_from_json(
                json.loads(self.i_snapshot_str.data)
            )
        else:
            self.o_snapshot_spec.data = SnapshotSpec.content_from_json(
                json.load(open(self.i_snapshot_file.data))
            )


class ContainerImagesFromSnapshot(Traction):
    """Extract container image references from snapshot."""

    i_snapshot_spec: In[SnapshotSpec]
    o_container_images: Out[TList[str]]

    d_: str = """Extract container image references from snapshot."""
    d_i_snapshot_spec: str = "Snapshot object"
    d_o_container_images: str = "List of container image references"

    def _run(self, on_update: OnUpdateCallable = None) -> None:
        for component in self.i_snapshot_spec.data.components:
            self.o_container_images.data.append(component.containerImage)



from unittest import mock
import pytest

from signtractions.resources.signing_wrapper import (
    SignerWrapper,
    SignerWrapperSettings,
    SignEntry,
    MsgSignerWrapper,
    MsgSignerSettings,
    SigningError,
)


def test_signer_wrapper_run_entrypoint():
    sw = SignerWrapper(config_file="", settings=SignerWrapperSettings())
    m = mock.Mock()
    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        mock_load_entry_point.return_value = m
        assert sw.entry_point == m
        # property test
        assert sw.entry_point == m


def test_signer_wrapper_remove_signatures():
    with mock.patch(
        "signtractions.resources.signing_wrapper.SignerWrapper._run_remove_signatures"
    ) as mocked_run_remove_signatures:
        sw = SignerWrapper(config_file="", settings=SignerWrapperSettings())
        sw.remove_signatures("test")
        mocked_run_remove_signatures.assert_called_once_with("test")


def test_signer_wrapper_run_store_signed():
    with mock.patch(
        "signtractions.resources.signing_wrapper.SignerWrapper._run_store_signed"
    ) as mocked_run_store_signed:
        sw = SignerWrapper(config_file="", settings=SignerWrapperSettings())
        m = mock.Mock()
        sw._store_signed(m)
        mocked_run_store_signed.assert_called_once_with(m)


def test_signer_wrapper_sign_containers():
    sw = SignerWrapper(config_file="", settings=SignerWrapperSettings())
    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        ep = mock.Mock()
        mock_load_entry_point.return_value = ep
        ep.return_value = {"signer_result": {"status": "ok"}}
        sw.sign_containers(
            [
                SignEntry(
                    repo="containers/podman",
                    reference="quay.io/containers/podman:latest",
                    digest="sha256:123456",
                    arch="amd64",
                    signing_key="signing_key",
                )
            ]
        )
        ep.assert_called_with(
            config_file="",
            signing_key="signing_key",
            reference=["quay.io/containers/podman:latest"],
            digest=["sha256:123456"],
        )


def test_signer_wrapper_sign_containers_error():
    sw = SignerWrapper(config_file="", settings=SignerWrapperSettings())
    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        ep = mock.Mock()
        mock_load_entry_point.return_value = ep
        ep.return_value = {"signer_result": {"status": "error", "error_message": "test_message"}}
        with pytest.raises(SigningError):
            sw.sign_containers(
                [
                    SignEntry(
                        repo="containers/podman",
                        reference="quay.io/containers/podman:latest",
                        digest="sha256:123456",
                        arch="amd64",
                        signing_key="signing_key",
                    )
                ]
            )


def test_signer_wrapper_sign_containers_nothing_to_sign():
    sw = SignerWrapper(config_file="", settings=SignerWrapperSettings())
    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        ep = mock.Mock()
        mock_load_entry_point.return_value = ep
        ep.return_value = {"signer_result": {"status": "ok"}}
        sw.sign_containers([])
        ep.assert_not_called()


def test_msg_signer_wrapper_sign_containers():
    sw = MsgSignerWrapper(
        config_file="",
        settings=MsgSignerSettings(pyxis_server="", pyxis_ssl_crt_file="", pyxis_ssl_key_file=""),
    )

    with mock.patch("pkg_resources.load_entry_point") as mock_load_entry_point:
        ep = mock.Mock()
        mock_load_entry_point.return_value = ep
        ep.return_value = {"signer_result": {"status": "ok"}}
        sw.sign_containers(
            [
                SignEntry(
                    repo="containers/podman",
                    reference="quay.io/containers/podman:latest",
                    digest="sha256:123456",
                    arch="amd64",
                    signing_key="signing_key",
                )
            ]
        )
        ep.assert_called_with(
            config_file="",
            signing_key="signing_key",
            reference=["quay.io/containers/podman:latest"],
            digest=["sha256:123456"],
        )


def test_msg_signer_wrapper_filter_to_sign():
    msw = MsgSignerWrapper(
        config_file="",
        settings=MsgSignerSettings(pyxis_server="", pyxis_ssl_crt_file="", pyxis_ssl_key_file=""),
    )
    with mock.patch(
        "signtractions.resources.signing_wrapper.MsgSignerWrapper._fetch_signatures"
    ) as mocked_fetch_signatures:
        mocked_fetch_signatures.return_value = []
        assert msw._filter_to_sign(
            [
                SignEntry(
                    repo="containers/podman",
                    reference="quay.io/containers/podman:latest",
                    digest="sha256:123456",
                    arch="amd64",
                    signing_key="signing_key",
                )
            ]
        ) == [
            SignEntry(
                repo="containers/podman",
                reference="quay.io/containers/podman:latest",
                digest="sha256:123456",
                arch="amd64",
                signing_key="signing_key",
            )
        ]


def test_msg_signer_wrapper_store_signed():
    sw = MsgSignerWrapper(
        config_file="",
        settings=MsgSignerSettings(pyxis_server="", pyxis_ssl_crt_file="", pyxis_ssl_key_file=""),
    )
    with mock.patch(
        "signtractions.resources.signing_wrapper.run_entrypoint_mod"
    ) as mock_run_entrypoint_mod:
        sw._store_signed(
            {
                "operation": {"references": ["registry.com/namespace/repo:latest"]},
                "signing_key": "signing_key",
                "operation_results": [
                    (
                        {
                            "msg": {
                                "repo": "repository",
                                "signed_claim": "signed_claim",
                                "manifest_digest": "digest",
                            },
                            "status": "ok",
                        },
                        True,
                    )
                ],
            }
        )
        mock_run_entrypoint_mod.assert_called_with(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-upload-signatures"),
            "pubtools-pyxis-upload-signature",
            [
                "--pyxis-server",
                "",
                "--pyxis-ssl-crtfile",
                "",
                "--pyxis-ssl-keyfile",
                "",
                "--request-threads",
                "7",
                "--signatures",
                mock.ANY,
            ],
            {},
        )


def test_msg_signer_wrapper_filter_to_sign_nothing_to_sign():
    msw = MsgSignerWrapper(
        config_file="",
        settings=MsgSignerSettings(pyxis_server="", pyxis_ssl_crt_file="", pyxis_ssl_key_file=""),
    )
    with mock.patch(
        "signtractions.resources.signing_wrapper.MsgSignerWrapper._fetch_signatures"
    ) as mocked_fetch_signatures:
        mocked_fetch_signatures.return_value = [
            {
                "repo": "containers/podman",
                "reference": "quay.io/containers/podman:latest",
                "manifest_digest": "sha256:123456",
                "arch": "amd64",
                "sig_key_id": "signing_key",
            }
        ]
        assert (
            msw._filter_to_sign(
                [
                    SignEntry(
                        repo="containers/podman",
                        reference="quay.io/containers/podman:latest",
                        digest="sha256:123456",
                        arch="amd64",
                        signing_key="signing_key",
                    )
                ]
            )
            == []
        )


def test_msg_signer_fetch_signature():
    with mock.patch(
        "signtractions.resources.signing_wrapper.run_entrypoint"
    ) as mocked_run_entrypoint:
        mocked_run_entrypoint.return_value = ["digest-1"]
        msw = MsgSignerWrapper(
            config_file="",
            settings=MsgSignerSettings(
                pyxis_server="", pyxis_ssl_crt_file="", pyxis_ssl_key_file=""
            ),
        )
        next(msw._fetch_signatures(["digest"]))
        mocked_run_entrypoint.assert_called_once_with(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-signatures"),
            "pubtools-pyxis-get-signatures",
            [
                "--pyxis-server",
                "",
                "--pyxis-ssl-crtfile",
                "",
                "--pyxis-ssl-keyfile",
                "",
                "--manifest-digest",
                mock.ANY,
            ],
            {},
        )


def test_msg_signer_remove_signature():
    with mock.patch(
        "signtractions.resources.signing_wrapper.run_entrypoint"
    ) as mocked_run_entrypoint:
        mocked_run_entrypoint.return_value = [
            {
                "manifest_digest": "digest",
                "reference": "registry.com/namespace/repo:latest",
                "repository": "repository",
                "_id": "id1",
                "sig_key_id": "signing_key",
            }
        ]
        msw = MsgSignerWrapper(
            config_file="",
            settings=MsgSignerSettings(
                pyxis_server="", pyxis_ssl_crt_file="", pyxis_ssl_key_file=""
            ),
        )
        msw.remove_signatures([("digest", "latest", "repository")])
        mocked_run_entrypoint.assert_has_calls(
            [
                mock.call(
                    ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-signatures"),
                    "pubtools-pyxis-get-signatures",
                    [
                        "--pyxis-server",
                        "",
                        "--pyxis-ssl-crtfile",
                        "",
                        "--pyxis-ssl-keyfile",
                        "",
                        "--manifest-digest",
                        mock.ANY,
                    ],
                    {},
                )
            ]
        )



from pytractions.base import TList
from signtractions.models.containers import ContainerParts


def test_container_parts():
    cp = ContainerParts(
        registry="registry",
        image="namespace/repo",
        tag="latest",
        digests=TList[str](["digest"]),
        arches=TList[str](["arch"]),
    )
    cp.make_reference() == "registry/namespace/repo:latest"

    cp = ContainerParts(
        registry="registry",
        image="namespace/repo",
        tag="",
        digests=TList[str](["digest"]),
        arches=TList[str](["arch"]),
    )
    cp.make_reference() == "registry/namespace/repo@digest"
